# -*- coding: utf-8 -*-
# Copyright (c) St. Anne's University Hospital in Brno. International Clinical
# Research Center, Biomedical Engineering. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.


# Std imports

# Third pary imports
import numpy as np
import pandas as pd

# Local imports


def calculate_absolute_samples(computation_result, window_size, overlap):
    """
    Convenience function to calculate absolute samples for the output of
    run_windowed class method.

    Parameters
    ----------
    computation_result: numpy.ndarray
        Output of run_windowed method
    window_size: int
        Size of the window in samples
    overlap: float
        Value between 0 and 1 specifying the fraction of the window to overlap

    Returns
    -------
    absolute_samples: numpy.ndarray
        Structured array with window starts
    """

    if overlap is None:
        overlap = 0

    return np.floor(computation_result['win_idx']
                    * window_size * (1 - overlap)).astype(np.int32)


def create_output_df(fields={}):
    """
    Function to create a custom pandas dataframe depending on the algorithm
    needs. Fields: event_start,event_stop are preset.

    Parameters
    ----------
    fields: dict
        Dictionary with keys as field names and values as data types

    Returns
    -------
    output_df: pandas dataframe
        Pandas dataframe with specified fields and field dtypes
    """

    dtype_dict = {'event_start': np.int32,
                  'event_stop': np.int32}

    dtype_dict.update(fields)

    out_df = pd.DataFrame(columns=dtype_dict.keys())
    out_df = out_df.astype(dtype=dtype_dict)

    return out_df


def add_metadata(df, metadata):
    """
    Convenience function to add metadata to the output dataframe.

    Parameters
    ----------
    df: pandas dataframe
        Dataframe with original data
    metadata: dict
        Dictionary with column_name:value

    Returns
    -------
    new_df: pandas dataframe
        Updated dataframe
    """

    for key in metadata.keys():
        df[key] = metadata[key]

    return df


def process_matrix(matrix, value, threshold_func, threshold_val):
    """
    Function for postprocessing of results generated by singal processing
    algorithms. The aim of the function is to detect outliers spanning
    accross multiple channels (usually artifacts).
    
    Parameters
    ----------
    matrix : np.array
        2D structured numpy array (channels X samples) of concacenated
        univariate or bivariate processing results. Alternativelly binned
        event counts.
    value : str
        Value string for the field in the structured array accross which the
        aggregation function will operate.
    threshold_func : function object
        Function for calculating 
    threshold_val : float
        Threshold for the threshold function.

    Returns
    -------
    count_sig : np.array
        1D numpy array with normlized (0-1) counts of values above the
        threshold accross channels.
        
    Examples
    --------
    
    from epycom.univariate import SignalStats
    
    signal_stats = SignalStats()
    
    res_list = []
    for data in channels_data:    
        res_list.append(signal_stats.run_windowed(data, 5000, 0, None))
    
    matrix = np.c_[res_list]
    
    count_sig = process_matrix(matrix, 'power_max', th_percentile, 95)
    """
        
    count_sig = np.zeros(matrix.shape[1])
    for i in range(matrix.shape[0]):
        th = threshold_func(matrix[value][i],threshold_val)
        th_wins = matrix['win_idx'][0][matrix[value][i]>th]
        count_sig[th_wins] += 1
    
    count_sig /= matrix.shape[0]
    
    return count_sig
